@@ -45,9 +45,11 @@ logger = logging.getLogger(__name__)
 
 # For the time being. This is JAR specific and should probably be configurable.
 ALG_PARAMS = {
-    "sign": ["request_object_signing_alg", "request_object_signing_alg_values_supported",],
-    "enc_alg": ["request_object_encryption_alg", "request_object_encryption_alg_values_supported",],
-    "enc_enc": ["request_object_encryption_enc", "request_object_encryption_enc_values_supported",],
+    "sign": ["request_object_signing_alg", "request_object_signing_alg_values_supported", ],
+    "enc_alg": ["request_object_encryption_alg",
+                "request_object_encryption_alg_values_supported", ],
+    "enc_enc": ["request_object_encryption_enc",
+                "request_object_encryption_enc_values_supported", ],
 }
 
 FORM_POST = """<html>
@@ -69,6 +71,8 @@ def inputs(form_args):
     element = []
     html_field = '<input type="hidden" name="{}" value="{}"/>'
     for name, value in form_args.items():
+        if name == "scope" and isinstance(value, list):
+            value = " ".join(value)
         element.append(html_field.format(name, value))
     return "\n".join(element)
 
@@ -79,10 +83,10 @@ def max_age(request):
 
 
 def verify_uri(
-    endpoint_context: EndpointContext,
-    request: Union[dict, Message],
-    uri_type: str,
-    client_id: Optional[str] = None,
+        endpoint_context: EndpointContext,
+        request: Union[dict, Message],
+        uri_type: str,
+        client_id: Optional[str] = None,
 ):
     """
     A redirect URI
@@ -100,8 +104,6 @@ def verify_uri(
     if not _cid:
         logger.error("No client id found")
         raise UnknownClient("No client_id provided")
-    else:
-        logger.debug("Client ID: {}".format(_cid))
 
     _uri = request.get(uri_type)
     if _uri is None:
@@ -122,7 +124,6 @@ def verify_uri(
     if client_info is None:
         raise KeyError("No such client")
 
-    logger.debug("Client info: {}".format(client_info))
     redirect_uris = client_info.get("{}s".format(uri_type))
     if redirect_uris is None:
         raise ValueError(f"No registered {uri_type} for {_cid}")
@@ -207,7 +208,7 @@ def get_uri(endpoint_context, request, uri_type):
 
 
 def authn_args_gather(
-    request: Union[AuthorizationRequest, dict], authn_class_ref: str, cinfo: dict, **kwargs,
+        request: Union[AuthorizationRequest, dict], authn_class_ref: str, cinfo: dict, **kwargs,
 ):
     """
     Gather information to be used by the authentication method
@@ -245,16 +246,20 @@ def authn_args_gather(
     return authn_args
 
 
-def check_unknown_scopes_policy(request_info, cinfo, endpoint_context):
-    op_capabilities = endpoint_context.conf["capabilities"]
-    client_allowed_scopes = cinfo.get("allowed_scopes") or op_capabilities["scopes_supported"]
+def check_unknown_scopes_policy(request_info, client_id, endpoint_context):
+    if not endpoint_context.conf["capabilities"].get("deny_unknown_scopes"):
+        return
 
+    scope = request_info["scope"]
+    filtered_scopes = set(
+        endpoint_context.scopes_handler.filter_scopes(scope, client_id=client_id)
+    )
+    scopes = set(scope)
     # this prevents that authz would be released for unavailable scopes
-    for scope in request_info["scope"]:
-        if op_capabilities.get("deny_unknown_scopes") and scope not in client_allowed_scopes:
-            _msg = "{} requested an unauthorized scope ({})"
-            logger.warning(_msg.format(cinfo["client_id"], scope))
-            raise UnAuthorizedClientScope()
+    if scopes != filtered_scopes:
+        diff = " ".join(scopes - filtered_scopes)
+        logger.warning(f"{client_id} requested unauthorized scopes: {diff}")
+        raise UnAuthorizedClientScope()
 
 
 class Authorization(Endpoint):
@@ -291,6 +296,13 @@ class Authorization(Endpoint):
     def extra_response_args(self, aresp):
         return aresp
 
+    def authentication_error_response(self, request, error, error_description, **kwargs):
+        _error_msg = self.error_cls(error=error, error_description=error_description)
+        _state = request.get("state")
+        if _state:
+            _error_msg["state"] = _state
+        return _error_msg
+
     def verify_response_type(self, request: Union[Message, dict], cinfo: dict) -> bool:
         # Checking response types
         _registered = [set(rt.split(" ")) for rt in cinfo.get("response_types", [])]
@@ -392,20 +404,23 @@ class Authorization(Endpoint):
         """
         if not request:
             logger.debug("No AuthzRequest")
-            return self.error_cls(
-                error="invalid_request", error_description="Can not parse AuthzRequest"
-            )
+            return self.authentication_error_response(request,
+                                                      error="invalid_request",
+                                                      error_description="Can not parse AuthzRequest"
+                                                      )
 
         request = self.filter_request(endpoint_context, request)
 
         _cinfo = endpoint_context.cdb.get(client_id)
         if not _cinfo:
             logger.error("Client ID ({}) not in client database".format(request["client_id"]))
-            return self.error_cls(error="unauthorized_client", error_description="unknown client")
+            return self.authentication_error_response(request, error="unauthorized_client",
+                                                      error_description="unknown client")
 
         # Is the asked for response_type among those that are permitted
         if not self.verify_response_type(request, _cinfo):
-            return self.error_cls(
+            return self.authentication_error_response(
+                request,
                 error="invalid_request",
                 error_description="Trying to use unregistered response_type",
             )
@@ -414,7 +429,8 @@ class Authorization(Endpoint):
         try:
             redirect_uri = get_uri(endpoint_context, request, "redirect_uri")
         except (RedirectURIError, ParameterError) as err:
-            return self.error_cls(
+            return self.authentication_error_response(
+                request,
                 error="invalid_request",
                 error_description="{}:{}".format(err.__class__.__name__, err),
             )
@@ -452,7 +468,7 @@ class Authorization(Endpoint):
     def create_session(self, request, user_id, acr, time_stamp, authn_method):
         _context = self.server_get("endpoint_context")
         _mngr = _context.session_manager
-        authn_event = create_authn_event(user_id, authn_info=acr, time_stamp=time_stamp,)
+        authn_event = create_authn_event(user_id, authn_info=acr, time_stamp=time_stamp, )
         _exp_in = authn_method.kwargs.get("expires_in")
         if _exp_in and "valid_until" in authn_event:
             authn_event["valid_until"] = utc_time_sans_frac() + _exp_in
@@ -466,14 +482,26 @@ class Authorization(Endpoint):
             token_usage_rules=_token_usage_rules,
         )
 
+    def _login_required_error(self, redirect_uri, request):
+        _res = {
+            "error": "login_required",
+            "return_uri": redirect_uri,
+            "return_type": request["response_type"],
+        }
+        _state = request.get("state")
+        if _state:
+            _res["state"] = _state
+        logger.debug("Login required error: {}".format(_res))
+        return _res
+
     def setup_auth(
-        self,
-        request: Optional[Union[Message, dict]],
-        redirect_uri: str,
-        cinfo: dict,
-        cookie: List[dict] = None,
-        acr: str = None,
-        **kwargs,
+            self,
+            request: Optional[Union[Message, dict]],
+            redirect_uri: str,
+            cinfo: dict,
+            cookie: List[dict] = None,
+            acr: str = None,
+            **kwargs,
     ):
         """
 
@@ -499,6 +527,7 @@ class Authorization(Endpoint):
                 _max_age = 0
             else:
                 _max_age = max_age(request)
+            logger.debug(f'Max age: {_max_age}')
             identity, _ts = authn.authenticated_as(
                 client_id, cookie, authorization=_auth_info, max_age=_max_age
             )
@@ -541,11 +570,7 @@ class Authorization(Endpoint):
 
             if "prompt" in request and "none" in request["prompt"]:
                 # Need to authenticate but not allowed
-                return {
-                    "error": "login_required",
-                    "return_uri": redirect_uri,
-                    "return_type": request["response_type"],
-                }
+                return self._login_required_error(redirect_uri, request)
             else:
                 return {"function": authn, "args": authn_args}
         else:
@@ -560,11 +585,7 @@ class Authorization(Endpoint):
                     if user != kwargs["req_user"]:
                         logger.debug("Wanted to be someone else!")
                         if "prompt" in request and "none" in request["prompt"]:
-                            # Need to authenticate but not allowed
-                            return {
-                                "error": "login_required",
-                                "return_uri": redirect_uri,
-                            }
+                            return self._login_required_error(redirect_uri, request)
                         else:
                             return {"function": authn, "args": authn_args}
 
@@ -605,12 +626,12 @@ class Authorization(Endpoint):
         return ""
 
     def response_mode(
-        self,
-        request: Union[dict, AuthorizationRequest],
-        response_args: Optional[AuthorizationResponse] = None,
-        return_uri: Optional[str] = "",
-        fragment_enc: Optional[bool] = None,
-        **kwargs,
+            self,
+            request: Union[dict, AuthorizationRequest],
+            response_args: Optional[AuthorizationResponse] = None,
+            return_uri: Optional[str] = "",
+            fragment_enc: Optional[bool] = None,
+            **kwargs,
     ) -> dict:
         resp_mode = request["response_mode"]
         if resp_mode == "form_post":
@@ -618,9 +639,9 @@ class Authorization(Endpoint):
                 _args = response_args.to_dict()
             else:
                 _args = response_args
-            msg = FORM_POST.format(inputs=inputs(_args), action=return_uri,)
+            msg = FORM_POST.format(inputs=inputs(_args), action=return_uri, )
             kwargs.update(
-                {"response_msg": msg, "content_type": "text/html", "response_placement": "body",}
+                {"response_msg": msg, "content_type": "text/html", "response_placement": "body", }
             )
         elif resp_mode == "fragment":
             if fragment_enc is False:
@@ -640,8 +661,10 @@ class Authorization(Endpoint):
 
         return kwargs
 
-    def error_response(self, response_info, error, error_description):
-        resp = self.error_cls(error=error, error_description=str(error_description))
+    def error_response(self, response_info, request, error, error_description):
+        resp = self.authentication_error_response(request,
+                                                  error=error,
+                                                  error_description=str(error_description))
         response_info["response_args"] = resp
         return response_info
 
@@ -664,7 +687,9 @@ class Authorization(Endpoint):
             _sinfo = _mngr.get_session_info(sid, grant=True)
 
             if request.get("scope"):
-                aresp["scope"] = request["scope"]
+                aresp["scope"] = _context.scopes_handler.filter_scopes(
+                    request["scope"], _sinfo["client_id"]
+                )
 
             rtype = set(request["response_type"][:])
             handled_response_type = []
@@ -715,7 +740,8 @@ class Authorization(Endpoint):
                     # id_token = _context.idtoken.make(sid, **kwargs)
                 except (JWEException, NoSuitableSigningKeys) as err:
                     logger.warning(str(err))
-                    resp = self.error_cls(
+                    resp = self.authentication_error_response(
+                        request,
                         error="invalid_request",
                         error_description="Could not sign/encrypt id_token",
                     )
@@ -726,7 +752,8 @@ class Authorization(Endpoint):
 
             not_handled = rtype.difference(handled_response_type)
             if not_handled:
-                resp = self.error_cls(
+                resp = self.authentication_error_response(
+                    request,
                     error="invalid_request", error_description="unsupported_response_type",
                 )
                 return {"response_args": resp, "fragment_enc": fragment_enc}
@@ -753,13 +780,14 @@ class Authorization(Endpoint):
 
         grant = _context.authz(session_id, request=request)
         if grant.is_active() is False:
-            return self.error_response(response_info, "server_error", "Grant not usable")
+            return self.error_response(response_info, request, "server_error", "Grant not usable")
 
         user_id, client_id, grant_id = _mngr.decrypt_session_id(session_id)
         try:
             _mngr.set([user_id, client_id, grant_id], grant)
         except Exception as err:
-            return self.error_response(response_info, "server_error", "{}".format(err.args))
+            return self.error_response(response_info, request, "server_error",
+                                       "{}".format(err.args))
 
         logger.debug("response type: %s" % request["response_type"])
 
@@ -771,7 +799,8 @@ class Authorization(Endpoint):
         try:
             redirect_uri = get_uri(_context, request, "redirect_uri")
         except (RedirectURIError, ParameterError) as err:
-            return self.error_response(response_info, "invalid_request", "{}".format(err.args))
+            return self.error_response(response_info, request, "invalid_request",
+                                       "{}".format(err.args))
         else:
             response_info["return_uri"] = redirect_uri
 
@@ -783,7 +812,8 @@ class Authorization(Endpoint):
             try:
                 response_info = self.response_mode(request, **response_info)
             except InvalidRequest as err:
-                return self.error_response(response_info, "invalid_request", "{}".format(err.args))
+                return self.error_response(response_info, request, "invalid_request",
+                                           "{}".format(err.args))
 
         _cookie_info = _context.new_cookie(
             name=_context.cookie_handler.name["session"],
@@ -807,7 +837,7 @@ class Authorization(Endpoint):
         try:
             resp_info = self.post_authentication(request, session_id, **kwargs)
         except Exception as err:
-            return self.error_response({}, "server_error", err)
+            return self.error_response({}, request, "server_error", err)
 
         _context = self.server_get("endpoint_context")
 
@@ -816,10 +846,11 @@ class Authorization(Endpoint):
             try:
                 authn_event = _context.session_manager.get_authentication_event(session_id)
             except KeyError:
-                return self.error_response({}, "server_error", "No such session")
+                return self.error_response({}, request, "server_error", "No such session")
             else:
                 if authn_event.is_valid() is False:
-                    return self.error_response({}, "server_error", "Authentication has timed out")
+                    return self.error_response({}, request, "server_error",
+                                               "Authentication has timed out")
 
             _state = b64e(as_bytes(json.dumps({"authn_time": authn_event["authn_time"]})))
 
@@ -850,8 +881,9 @@ class Authorization(Endpoint):
             resp_info["response_args"]["session_state"] = _session_state
 
         # Mix-Up mitigation
-        resp_info["response_args"]["iss"] = _context.issuer
-        resp_info["response_args"]["client_id"] = request["client_id"]
+        if "response_args" in resp_info:
+            resp_info["response_args"]["iss"] = _context.issuer
+            resp_info["response_args"]["client_id"] = request["client_id"]
 
         return resp_info
 
@@ -859,10 +891,10 @@ class Authorization(Endpoint):
         return kwargs
 
     def process_request(
-        self,
-        request: Optional[Union[Message, dict]] = None,
-        http_info: Optional[dict] = None,
-        **kwargs,
+            self,
+            request: Optional[Union[Message, dict]] = None,
+            http_info: Optional[dict] = None,
+            **kwargs,
     ):
         """ The AuthorizationRequest endpoint
 
@@ -877,22 +909,25 @@ class Authorization(Endpoint):
         _cid = request["client_id"]
         _context = self.server_get("endpoint_context")
         cinfo = _context.cdb[_cid]
-        logger.debug("client {}: {}".format(_cid, cinfo))
+        # logger.debug("client {}: {}".format(_cid, cinfo))
 
         # this apply the default optionally deny_unknown_scopes policy
-        if cinfo:
-            check_unknown_scopes_policy(request, cinfo, _context)
+        check_unknown_scopes_policy(request, _cid, _context)
 
         if http_info is None:
             http_info = {}
 
         _cookies = http_info.get("cookie")
         if _cookies:
-            _cookies = _context.cookie_handler.parse_cookie("oidcop", _cookies)
+            logger.debug("parse_cookie@process_request")
+            _session_cookie_name = _context.cookie_handler.name["session"]
+            _my_cookies = _context.cookie_handler.parse_cookie(_session_cookie_name, _cookies)
+        else:
+            _my_cookies = {}
 
         kwargs = self.do_request_user(request_info=request, **kwargs)
 
-        info = self.setup_auth(request, request["redirect_uri"], cinfo, _cookies, **kwargs)
+        info = self.setup_auth(request, request["redirect_uri"], cinfo, _my_cookies, **kwargs)
 
         if "error" in info:
             return info
@@ -901,7 +936,7 @@ class Authorization(Endpoint):
         if not _function:
             logger.debug("- authenticated -")
             logger.debug("AREQ keys: %s" % request.keys())
-            return self.authz_part2(request=request, cookie=_cookies, **info)
+            return self.authz_part2(request=request, cookie=_my_cookies, **info)
 
         try:
             # Run the authentication function
@@ -934,11 +969,19 @@ class AllowedAlgorithms:
 
 def re_authenticate(request, authn) -> bool:
     """
-    This is where you can demand reauthentication even though the authentication in use
+    This is where you can demand re-authentication even though the authentication in use
     is still valid.
 
     :param request:
     :param authn:
     :return:
     """
+    logger.debug("Re-authenticate ??: {}".format(request))
+
+    _prompt = request.get("prompt", [])
+    logger.debug(f"Prompt={_prompt}")
+    if "login" in _prompt:
+        logger.debug("Reauthenticate due to prompt=login")
+        return True
+
     return False
