@@ -7,12 +7,10 @@ from cryptojwt.jws.exception import JWSException
 from oidcop.exception import ToOld
 from oidcop.token import Crypt
 from oidcop.token.exception import WrongTokenClass
-
 from . import Token
 from . import is_expired
 from .exception import UnknownToken
-
-# TYPE_MAP = {"A": "code", "T": "access_token", "R": "refresh_token"}
+from ..constant import DEFAULT_TOKEN_LIFETIME
 
 
 class JWTToken(Token):
@@ -23,7 +21,7 @@ class JWTToken(Token):
             issuer: str = None,
             aud: Optional[list] = None,
             alg: str = "ES256",
-            lifetime: int = 300,
+            lifetime: int = DEFAULT_TOKEN_LIFETIME,
             server_get: Callable = None,
             token_type: str = "Bearer",
             password: str = "",
@@ -48,8 +46,13 @@ class JWTToken(Token):
         # inherit me and do your things here
         return payload
 
-    def __call__(self, session_id: Optional[str] = "", token_class: Optional[str] = "",
-                 **payload) -> str:
+    def __call__(
+        self,
+        session_id: Optional[str] = "",
+        token_class: Optional[str] = "",
+        usage_rules: Optional[dict] = None,
+        **payload
+    ) -> str:
 
         """
         Return a token.
@@ -70,8 +73,15 @@ class JWTToken(Token):
 
         # payload.update(kwargs)
         _context = self.server_get("endpoint_context")
+        if usage_rules and "expires_in" in usage_rules:
+            lifetime = usage_rules.get("expires_in")
+        else:
+            lifetime = self.lifetime
         signer = JWT(
-            key_jar=_context.keyjar, iss=self.issuer, lifetime=self.lifetime, sign_alg=self.alg,
+            key_jar=_context.keyjar,
+            iss=self.issuer,
+            lifetime=lifetime,
+            sign_alg=self.alg,
         )
 
         return signer.pack(payload)
