@@ -5,12 +5,13 @@ import json
 import logging
 import sys
 import time
-import warnings
 from typing import List
 from urllib.parse import unquote
+import warnings
 
 from cryptography.hazmat.primitives.ciphers.aead import AESGCM
 from cryptojwt.jwt import JWT
+from oidcmsg.time_util import utc_time_sans_frac
 
 from oidcop.exception import FailedAuthentication
 from oidcop.exception import ImproperlyConfigured
@@ -65,6 +66,15 @@ class UserAuthnMethod(object):
             return None, 0
         else:
             _info = self.cookie_info(cookie, client_id)
+            logger.debug('Cookie info: {}'.format(_info))
+            if _info:
+                if 'max_age' in kwargs and kwargs["max_age"] != 0:
+                    _max_age = kwargs["max_age"]
+                    _now = utc_time_sans_frac()
+                    if _now > _info["timestamp"] + _max_age:
+                        logger.debug(
+                            "Too old by {} seconds".format(_now - (_info["timestamp"] + _max_age)))
+                        return None, 0
             return _info, time.time()
 
     def verify(self, *args, **kwargs):
@@ -93,22 +103,30 @@ class UserAuthnMethod(object):
     def cookie_info(self, cookie: List[dict], client_id: str) -> dict:
         _context = self.server_get("endpoint_context")
         try:
+            logger.debug("parse_cookie@UserAuthnMethod")
             vals = _context.cookie_handler.parse_cookie(
                 cookies=cookie, name=_context.cookie_handler.name["session"]
             )
         except (InvalidCookieSign, AssertionError, AttributeError) as err:
             logger.warning(err)
-            vals = None
+            vals = []
+
+        logger.debug("Value cookies: {}".format(vals))
 
         if vals is None:
             pass
         else:
             for val in vals:
                 _info = json.loads(val["value"])
-                _, cid, _ = _context.session_manager.decrypt_session_id(_info["sid"])
-                if cid != client_id:
+                _info["timestamp"] = int(val["timestamp"])
+                session_id = _context.session_manager.decrypt_session_id(_info["sid"])
+                logger.debug("session id: {}".format(session_id))
+                # _, cid, _ = _context.session_manager.decrypt_session_id(_info["sid"])
+                if session_id[1] != client_id:
                     continue
                 else:
+                    _info["uid"] = session_id[0]
+                    _info["grant_id"] = session_id[2]
                     return _info
         return {}
 
@@ -130,13 +148,13 @@ class UserPassJinja2(UserAuthnMethod):
     url_endpoint = "/verify/user_pass_jinja"
 
     def __init__(
-        self,
-        db,
-        template_handler,
-        template="user_pass.jinja2",
-        server_get=None,
-        verify_endpoint="",
-        **kwargs,
+            self,
+            db,
+            template_handler,
+            template="user_pass.jinja2",
+            server_get=None,
+            verify_endpoint="",
+            **kwargs,
     ):
 
         super(UserPassJinja2, self).__init__(server_get=server_get)
